from altprint.printable.base import BasePrint
from altprint.layer import Layer
from altprint.gcode import GcodeExporter

class MultiProcess():
    def __init__(self, **kwargs):
        prop_defaults = {
            "parts": [],
            "gcode_exporter": GcodeExporter,
            "start_script": "",
            "end_script": "",
            "offset": (0,0,0),
            "verbose": True,
        }

        for (prop, default) in prop_defaults.items():
            setattr(self, prop, kwargs.get(prop, default))

class MultiPrint(BasePrint):

    _height = float
    _layers_dict = dict[_height, Layer]

    def __init__(self, process):
        self.process = process
        self.layers: _layers_dict = {}

    def slice(self):
        pass

    def make_layers(self):
        if self.process.verbose == True:
            print("Making the layers for the multipart ...")
        heights = []
        for part in self.process.parts:
            heights.extend(list(part.layers.keys()))

        heights = list(set(heights))
        heights.sort()

        for h in heights:
            layer = Layer(None, None, None, None, None)
            for part in self.process.parts:
                if h in part.layers.keys():
                    layer.infill.extend(part.layers[h].infill)
                    layer.perimeter.extend(part.layers[h].perimeter)
            self.layers[h] = layer

    def export_gcode(self, filename):
        if self.process.verbose == True:
            print("exporting gcode to {}".format(filename))
        gcode_exporter = self.process.gcode_exporter(start_script=self.process.start_script,
                                                     end_script=self.process.end_script)
        gcode_exporter.make_gcode(self)
        gcode_exporter.export_gcode(filename)



from shapely.geometry import Polygon, LineString, Point, MultiLineString, MultiPoint, GeometryCollection, LinearRing
from shapely.ops import linemerge, split
from shapely.affinity import rotate
from shapely import wkt
import numpy as np
from pulp import *
from altprint.infill.infill import InfillMethod
from altprint.layer import Layer, Raster

def get_bounds(polygon):
    minx = polygon.bounds[0]
    miny = polygon.bounds[1]
    maxx = polygon.bounds[2]
    maxy = polygon.bounds[3]
    return minx, miny, maxx, maxy


def get_hlines(polygon, gap):
    h_lines = []
    minx, miny, maxx, maxy = get_bounds(polygon)
    num_hlines = int((maxy - miny)/gap) + 1
    adj_gap = (maxy-miny)/num_hlines
    heights = []
    for i in range(num_hlines):
        heights.append(gap*i+miny)
    for h in heights:
        h_lines.append(LineString([(minx, h), (maxx, h)]))
    return h_lines


def get_intersections(polygon, hlines):
    intersections = []
    hlines_by_heights = []
    for i, line in enumerate(hlines):
        hlines_by_heights.append([])
        intersection = [polygon.intersection(line)]
        if type(intersection[0]) == GeometryCollection:
            intersection = list(intersection[0])
        for geom in intersection:
            if type(geom) == MultiLineString:
                merged = linemerge(geom)
                if type(merged) == LineString:
                    line = LineString([merged.coords[0], merged.coords[-1]])
                    intersections.append(line)
                    hlines_by_heights[i].append(line)
                else:
                    for line in merged.geoms:
                        intersections.append(line)
                        hlines_by_heights[i].append(line)
            elif type(geom) == LineString:
                intersections.append(geom)
                hlines_by_heights[i].append(line)

    return intersections, hlines_by_heights


def geom_to_list(geom): #multilinestring or linestring to list
    if type(geom) == LineString:
        return [geom]
    else:
        lines = list(geom.geoms)
        return lines

def remove_same_height_lines(hlines, clines):
    filtered_lines = []
    for line in clines:
        a = Point(line.coords[0])
        b = Point(line.coords[-1])
        if not a.coords[0][1] == b.coords[0][1]:
            for hline in hlines:
                if a.touches(hline):
                    for hline2 in hlines:
                        if b.touches(hline2):
                            filtered_lines.append(line)
                            break


    return filtered_lines

def round_coords(lines, precision):
    rounded_lines = []
    for line in lines:
        rounded_lines.append(wkt.loads(wkt.dumps(line, rounding_precision=precision)))
    return rounded_lines

def get_lines(shape, gap):
    hlines = get_hlines(shape, gap)
    lines, hlines_by_heights = get_intersections(shape, hlines)
    connection_lines = LinearRing(shape.exterior).difference(MultiLineString(hlines))
    connection_lines = geom_to_list(connection_lines)
    for hole in shape.interiors:
        extra_connection_lines = LinearRing(hole).difference(MultiLineString(hlines))
        extra_connection_lines = geom_to_list(extra_connection_lines)
        connection_lines.extend(extra_connection_lines)
    connection_lines = remove_same_height_lines(lines, connection_lines)

    precision = 6
    r_connection_lines = round_coords(connection_lines, precision)
    r_hlines_by_heights = []
    for height in hlines_by_heights:
        r_hlines_by_heights.append(round_coords(height, precision))
    return lines, connection_lines, hlines_by_heights

def get_connections_by_heights(clines, polygon, gap):
    minx, miny, maxx, maxy = get_bounds(polygon)
    num_hlines = int((maxy - miny)/gap) + 1
    edges = linemerge(MultiLineString(clines))

    clines_by_heights = []
    if type(edges) == LineString:
        edges = MultiLineString([edges])
    theta = np.ones((len(edges.geoms), num_hlines))
    for i, edge in enumerate(edges.geoms):
        clines_by_heights.append([])
        for j in range(num_hlines):
            clines_by_heights[i].append(None)
        for line in clines:
            if line.within(edge):
                j = int((line.bounds[1]-miny)/gap)
                clines_by_heights[i][j] = line
                theta[i][j] = 0
    return clines_by_heights, theta

def get_c(hlines_by_heights, clines_by_heights):
    c = np.zeros((len(clines_by_heights), len(clines_by_heights), len(clines_by_heights[0])))
    for i in range(len(c)):
        for j in range(len(c[0])):
            for k in range(len(c[0][0])):
                a = clines_by_heights[i][k]
                b = clines_by_heights[j][k]
                if a != None and b != None:
                    for hline in hlines_by_heights[k]:
                        if a.touches(hline) and b.touches(hline):
                            c[i][j][k] = 1
                            break
    return c

def get_x(clines):
    x = []
    for i in range(len(clines)):
        x.append([])
        for j in range(len(clines[i])):
            x[i].append("x{}.{}".format(i, j))
    return x

def get_all(shape, gap):
    hlines, pre_clines, hlines_by_heights = get_lines(shape, gap)
    clines, theta = get_connections_by_heights(pre_clines, shape, gap)
    if not clines:
        return None, None, None, None, None
    c = get_c(hlines_by_heights, clines)
    x = get_x(clines)
    return hlines, clines, theta, c, x

def rectilinear_optimal(shape, gap, angle):
    shape = rotate(shape, angle, origin=(0,0))
    hlines, clines, theta, c, x = get_all(shape, gap)
    if not clines:
        return []
    prob = LpProblem("Problema_Teste",LpMaximize)

    xf = [j for i in x for j in i]
    z = LpVariable.dicts("LC",xf,cat = 'Binary')

    prob += lpSum(z)

    '''restrição de não poder usar uma linha de conexão que não existe'''
    for i in range(len(x)):
        for j in range(len(x[i])-1):
            prob += z[x[i][j]] + z[x[i][j+1]] <= 1
            prob += theta[i][j] + z[x[i][j]] <= 1
        prob += theta[i][-1] + z[x[i][-1]] <= 1

    ''' restrição não poder ir para duas linhas de conexão que possuem uma ligação'''
    for i in range(len(c)):
        for j in range(i+1,len(c[i])):
            for k in range(len(c[i][j])):
                if(c[i][j][k] != 0):
                    prob += c[i][j][k] * (z[x[i][k]] + z[x[j][k]]) <= 1
    prob.solve(PULP_CBC_CMD(msg=0));
    tempo = prob.solutionTime
    num_restr = len(prob.constraints)
    num_var = len(prob._variables)

    vecSol = [v.varValue for v in prob.variables()]

    valor = value(prob.objective)

    selected_lines = []
    for v in prob.variables():
        if v.varValue == 1:
            i = int(v.name[4:].split(".")[0])
            j = int(v.name[4:].split(".")[1])
            selected_lines.append(clines[i][j])

    selected_lines.extend(hlines)
    paths = linemerge(MultiLineString(selected_lines))
    if type(paths) == LineString:
        return [paths]
    else:
        return list(paths.geoms)

class RectilinearOptimal(InfillMethod):

    def generate_infill(self, layer: Layer, gap, angle) -> MultiLineString:
        infill = []
        for border in layer.infill_border.geoms:
            infill.extend(rectilinear_optimal(border, gap, angle))
        return rotate(MultiLineString(infill), -angle, origin=(0,0))



from altprint.printable.base import BasePrint
from altprint.slicer import STLSlicer
from altprint.height_method import CopyHeightsFromFileMethod
from altprint.gcode import GcodeExporter


class InjectionProcess():
    def __init__(self, **kwargs):
        prop_defaults = {
            "parts": [],
            "parts_offset": [0,0,0],
            "source_gcode": '',
            "verbose": True,
        }

        for (prop, default) in prop_defaults.items():
            setattr(self, prop, kwargs.get(prop, default))

class InjectionPrint(BasePrint):
    """Base Printable Object"""

    def __init__(self, process: InjectionProcess):
        self.process = process
        self.layers: _layers_dict = {}
        self.layers_gcode = {}
        for part in self.process.parts:
            part.process.slicer = STLSlicer(CopyHeightsFromFileMethod(self.process.source_gcode))
            part.process.offset = self.process.parts_offset

    def slice(self):
        for part in self.process.parts:
            part.slice()

    def make_layers(self):
        for part in self.process.parts:
            part.make_layers()

    def make_layers_gcode(self):
        gcode_exporter = GcodeExporter()
        heights = CopyHeightsFromFileMethod(self.process.source_gcode).get_heights()
        for height in heights:
            layer_gcode = []
            for part in self.process.parts:
                layer_gcode.extend(gcode_exporter.make_layer_gcode(part.layers[height]))
            self.layers_gcode[height] = layer_gcode
        
    def export_gcode(self, filename):
        self.make_layers_gcode()
        output_lines = []
        with open(self.process.source_gcode, "r") as f:
            lines = f.readlines()
        for line in lines:
            output_lines.append(line)
            if line.startswith("; ALTPRINT"):
                height = float(line.split(' ')[-1])
                output_lines.extend(self.layers_gcode[height])
            
        with open(filename, "w") as f:
            for line in output_lines:
                f.write(line)


from altprint.printable.base import BasePrint
from altprint.slicer import STLSlicer
from altprint.layer import Layer, Raster
from altprint.height_method import StandartHeightMethod
from altprint.infill.rectilinear_optimal import RectilinearOptimal
from altprint.flow import calculate
from altprint.gcode import GcodeExporter
from altprint.lineutil import split_by_regions, retract
from altprint.settingsparser import SettingsParser

class FlexProcess():
    def __init__(self, **kwargs):
        prop_defaults = {
            "model_file": "",
            "flex_model_file": "",
            "slicer": STLSlicer(StandartHeightMethod()),
            "infill_method": RectilinearOptimal,
            "infill_angle": 0,
            "offset": (0, 0, 0),
            "external_adjust": 0.5,
            "perimeter_num": 1,
            "perimeter_gap": 0.5,
            "raster_gap": 0.5,
            "overlap": 0.0,
            "skirt_distance": 10,
            "skirt_num": 3,
            "skirt_gap": 0.5,
            "first_layer_flow": 2,
            "flow": 1.2,
            "speed": 2400,
            "flex_flow": 0,
            "flex_speed": 2000,
            "retract_flow": 2,
            "retract_speed": 1200,
            "retract_ratio": 0.9,
            "gcode_exporter": GcodeExporter,
            "start_script": "",
            "end_script": "",
            "verbose": True,
        }

        for (prop, default) in prop_defaults.items():
            setattr(self, prop, kwargs.get(prop, default))
        
        if 'settings_file' in kwargs.keys():
            settings = SettingsParser().load_from_file(kwargs['settings_file'])
            for (setting, value) in settings.items():
                setattr(self, setting, value)


class FlexPrint(BasePrint):
    """The common print. Nothing special"""

    _height = float
    _layers_dict = dict[_height, Layer]

    def __init__(self, process: FlexProcess):
        self.process = process
        self.layers: _layers_dict = {}
        self.heights: list[float] = []

    def slice(self):
        if self.process.verbose == True:
            print("slicing {} ...".format(self.process.model_file))
        slicer = self.process.slicer
        slicer.load_model(self.process.model_file)
        slicer.translate_model(self.process.offset)
        self.sliced_planes = slicer.slice_model()
        self.heights = self.sliced_planes.get_heights()

        slicer.load_model(self.process.flex_model_file)
        slicer.translate_model(self.process.offset)
        self.flex_planes = slicer.slice_model(self.heights)

    def make_layers(self):
        if self.process.verbose == True:
            print("generating layers ...")
        infill_method = self.process.infill_method()
        
        """skirt = Layer(self.sliced_planes.planes[self.heights[0]],
                      self.process.skirt_num,
                      self.process.skirt_gap,
                      - self.process.skirt_distance - self.process.skirt_gap * self.process.skirt_num,
                      self.process.overlap)
        skirt.make_perimeter()"""

        for i, height in enumerate(self.heights):
            layer = Layer(self.sliced_planes.planes[height],
                          self.process.perimeter_num,
                          self.process.perimeter_gap,
                          self.process.external_adjust,
                          self.process.overlap)
            if layer.shape == []:
                self.layers[height] = layer
                continue
            layer.make_perimeter()
            layer.make_infill_border()
            infill_paths = infill_method.generate_infill(layer,
                                                   self.process.raster_gap,
                                                   self.process.infill_angle)
            flex_regions = self.flex_planes.planes[height]

            if not type(flex_regions) == list:
                flex_regions = list(flex_regions.geoms)

            layer.perimeter_paths = split_by_regions(layer.perimeter_paths, flex_regions)
            infill_paths = split_by_regions(infill_paths, flex_regions)
            """if i==0: #skirt
                for path in skirt.perimeter_paths.geoms:
                    layer.perimeter.append(Raster(path, self.process.first_layer_flow, self.process.speed))"""
            for path in layer.perimeter_paths.geoms:
                flex_path = False
                for region in flex_regions:
                    if path.within(region.buffer(0.01, join_style=2)):
                        flex_path, retract_path = retract(path, self.process.retract_ratio)
                        layer.perimeter.append(Raster(flex_path, self.process.flex_flow, self.process.flex_speed))
                        layer.perimeter.append(Raster(retract_path, self.process.retract_flow, self.process.retract_speed))
                        flex_path = True
                        break
                if not flex_path:
                    if i==0: 
                        layer.perimeter.append(Raster(path, self.process.first_layer_flow, self.process.speed))
                    else:
                        layer.perimeter.append(Raster(path, self.process.flow, self.process.speed))

            for path in infill_paths.geoms:
                flex_path = False
                for region in flex_regions:
                    if path.within(region.buffer(0.01, join_style=2)):
                        flex_path, retract_path = retract(path, self.process.retract_ratio)
                        layer.infill.append(Raster(flex_path, self.process.flex_flow, self.process.flex_speed))
                        layer.infill.append(Raster(retract_path, self.process.retract_flow, self.process.retract_speed))
                        flex_path = True
                        break
                if not flex_path:
                    if i==0:
                        layer.infill.append(Raster(path, self.process.first_layer_flow, self.process.speed))
                    else:
                        layer.infill.append(Raster(path, self.process.flow, self.process.speed))
            self.layers[height] = layer

    def export_gcode(self, filename):
        if self.process.verbose == True:
            print("exporting gcode to {}".format(filename))

        gcode_exporter = self.process.gcode_exporter(start_script=self.process.start_script,
                                                     end_script=self.process.end_script)
        gcode_exporter.make_gcode(self)
        gcode_exporter.export_gcode(filename)



from shapely.ops import split
from shapely.geometry import LineString, MultiLineString

def retract(path, ratio):
    x, y = path.xy
    A = (x[0], y[0])
    B = (x[-1], y[-1])
    C = (A[0] + ratio*(B[0]-A[0]), A[1] + ratio*(B[1]-A[1]))
    flex_path = LineString([A, C])
    retract_path = LineString([C, B])
    return flex_path, retract_path

def split_lines(lines, spliter):
    final = []
    for line in lines:
        splited = split(line, spliter)
        for i in list(splited.geoms):
            if type(i) == LineString:
                final.append(i)
            else:
                print("not linestring")
    return final

def split_by_regions(lines, regions):
    final = list(lines.geoms)
    for region in regions:
        final = split_lines(final, region)
    return MultiLineString(final)



from abc import ABC, abstractmethod
import numpy as np


class HeightMethod(ABC):
    """Generates the height values on which the object will be sliced in"""

    @abstractmethod
    def get_heights(self, bounds) -> list[float]:
        pass

class StandartHeightMethod(HeightMethod):
    """Evenly spaced layers"""

    def __init__(self, layer_height: float = 0.2):
        self.layer_height = layer_height

    def get_heights(self, bounds) -> list[float]:
        zi = bounds[0][2] + self.layer_height
        zf = bounds[1][2]
        h = zf - zi
        heights = list(np.linspace(zi, zf, round(h/self.layer_height)+1))
        heights[-1] = heights[-1]-0.001 #numerical adjust to make the slicer include the last layer
        heights = list(np.around(heights, decimals=3))
        return heights


class CopyHeightsFromFileMethod(HeightMethod):
    """Get Heights from a premade gcode file"""

    def __init__(self, gcode_file_name: str):
        self.gcode_file_name = gcode_file_name

    def get_heights(self, bounds=None) -> list[float]:
        heights = []
        with open(self.gcode_file_name, "r") as f:
            lines = f.readlines()
        for line in lines:
            if line.startswith("; ALTPRINT"):
                heights.append(float(line.split(' ')[-1]))
        return heights

if __name__ == "__main__":
    cp = CopyHeightsFromFileMethod("teste.gcode")
    print(cp.get_heights())


from altprint.printable.base import BasePrint
from altprint.slicer import STLSlicer
from altprint.layer import Layer, Raster
from altprint.height_method import StandartHeightMethod
from altprint.infill.rectilinear_optimal import RectilinearOptimal
from altprint.gcode import GcodeExporter
from altprint.settingsparser import SettingsParser

class StandartProcess():
    def __init__(self, **kwargs):
        prop_defaults = {
            "model_file": "",
            "slicer": STLSlicer(StandartHeightMethod()),
            "infill_method": RectilinearOptimal,
            "infill_angle": [0, 90],
            "offset": (0, 0, 0),
            "external_adjust": 0.5,
            "perimeter_num": 2,
            "perimeter_gap": 0.5,
            "skirt_distance": 10,
            "skirt_num": 3,
            "skirt_gap": 0.5,
            "raster_gap": 0.5,
            "overlap": 0.0,
            "speed": 2400,
            "flow": 1.2,
            "gcode_exporter": GcodeExporter,
            "start_script": "",
            "end_script": "",
            "verbose": True,
        }


        for (prop, default) in prop_defaults.items():
            setattr(self, prop, kwargs.get(prop, default))

        if 'settings_file' in kwargs.keys():
            settings = SettingsParser().load_from_file(kwargs['settings_file'])
            for (setting, value) in settings.items():
                setattr(self, setting, value)

class StandartPrint(BasePrint):
    """The common print. Nothing special"""

    _height = float
    _layers_dict = dict[_height, Layer]

    def __init__(self, process: StandartProcess):
        self.process = process
        self.layers: _layers_dict = {}
        self.heights: list[float] = []

    def slice(self):
        if self.process.verbose == True:
            print("slicing {} ...".format(self.process.model_file))
        slicer = self.process.slicer
        slicer.load_model(self.process.model_file)
        slicer.translate_model(self.process.offset)
        self.sliced_planes = slicer.slice_model()
        self.heights = self.sliced_planes.get_heights()

    def make_layers(self):
        if self.process.verbose == True:
            print("generating layers ...")
        infill_method = self.process.infill_method()

        skirt = Layer(self.sliced_planes.planes[self.heights[0]],
                      self.process.skirt_num,
                      self.process.skirt_gap,
                      - self.process.skirt_distance - self.process.skirt_gap * self.process.skirt_num,
                      self.process.overlap)
        skirt.make_perimeter()

        for i, height in enumerate(self.heights):
            layer = Layer(self.sliced_planes.planes[height],
                          self.process.perimeter_num,
                          self.process.perimeter_gap,
                          self.process.external_adjust,
                          self.process.overlap)
            layer.make_perimeter()
            layer.make_infill_border()
            if type(self.process.infill_angle) == list:
                infill_angle = self.process.infill_angle[i%len(self.process.infill_angle)]
            else:
                infill_angle = self.process.infill_angle
            infill_paths = infill_method.generate_infill(layer,
                                                         self.process.raster_gap,
                                                         infill_angle)

            if i==0: #skirt
                for path in skirt.perimeter_paths.geoms:
                    layer.perimeter.append(Raster(path, self.process.flow, self.process.speed))

            for path in layer.perimeter_paths.geoms:
                layer.perimeter.append(Raster(path, self.process.flow, self.process.speed))
            for path in infill_paths.geoms:
                layer.infill.append(Raster(path, self.process.flow, self.process.speed))
            self.layers[height] = layer

    def export_gcode(self, filename):
        if self.process.verbose == True:
            print("exporting gcode to {}".format(filename))
        gcode_exporter = self.process.gcode_exporter(start_script=self.process.start_script,
                                                     end_script=self.process.end_script)
        gcode_exporter.make_gcode(self)
        gcode_exporter.export_gcode(filename)



from abc import ABC, abstractmethod
from shapely.geometry import MultiLineString
from altprint.layer import Layer

class InfillMethod(ABC):

    @abstractmethod
    def generate_infill(self, layer: Layer) -> MultiLineString:
        pass



from abc import ABC, abstractmethod

class BasePrint(ABC):
    """Base Printable Object"""

    @abstractmethod
    def slice(self):
        pass

    @abstractmethod
    def make_layers(self):
        pass

    @abstractmethod
    def export_gcode(self, filename):
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="altprint",
    version="0.2.0",
    author="couto0",
    author_email="daniel.couto64@gmail.com",
    description="Alternative 3d printing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/couto0/altprint",
    keywords='FFF FDM 3dprinting slicer slicing gcode',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['numpy>=1.19.4', 'Shapely>=1.7.1', 'trimesh>=3.9.1', 'PuLP>=2.4', 'scipy>=1.7.0', 'networkx>=2.5.1', 'rtree>=0.9.7', 'PyYAML>=6.0'],
)



from shapely.geometry import Polygon, MultiPolygon, LineString, MultiLineString
import numpy as np
from altprint.flow import calculate

class Raster:

    def __init__(self, path: LineString, flow, speed):

        self.path = path

        self.speed = np.ones(len(path.coords)) * speed
        self.extrusion = np.zeros(len(path.coords))
        x, y = path.xy
        for i in range(1, len(path.coords)):
            dx = abs(x[i] - x[i - 1])
            dy = abs(y[i] - y[i - 1])
            self.extrusion[i] = np.sqrt((dx**2) + (dy**2)) * flow * calculate() + self.extrusion[i-1]


class Layer:
    """Layer Object that stores layer internal and external shapes, also perimeters and infill path"""

    def __init__(self, shape: MultiPolygon, perimeter_num, perimeter_gap, external_adjust, overlap):
        self.shape = shape
        self.perimeter_num = perimeter_num
        self.perimeter_gap = perimeter_gap
        self.external_adjust = external_adjust
        self.overlap = overlap
        self.perimeter_paths: List = []
        self.perimeter: List = []
        self.infill: List = []
        self.infill_border: MultiPolygon = MultiPolygon()

    def make_perimeter(self):
        """Generates the perimeter based on the layer process"""

        perimeter_paths = []
        for section in self.shape.geoms:
            for i in range(self.perimeter_num):
                eroded_shape = section.buffer(- self.perimeter_gap*(i)
                                              - self.external_adjust/2, join_style=2)

                if eroded_shape.is_empty:
                    break
                if type(eroded_shape) == Polygon:
                    polygons = [eroded_shape]
                elif type(eroded_shape) == MultiPolygon:
                    polygons = list(eroded_shape.geoms)

                for poly in polygons:
                    for hole in poly.interiors:
                        perimeter_paths.append(LineString(hole))
                for poly in polygons:
                    perimeter_paths.append(LineString(poly.exterior))
        self.perimeter_paths = MultiLineString(perimeter_paths)

    def make_infill_border(self):
        """Generates the infill border based on the layer process"""

        infill_border_geoms = []
        for section in self.shape.geoms:
            eroded_shape = section.buffer(- self.perimeter_gap
                                          * self.perimeter_num
                                          - self.external_adjust/2
                                          + self.overlap, join_style=2)
            if not eroded_shape.is_empty:
                if type(eroded_shape) == Polygon:
                    infill_border_geoms.append(eroded_shape)
                else:
                    infill_border_geoms.extend(eroded_shape.geoms)

        self.infill_border = MultiPolygon(infill_border_geoms)



from abc import ABC, abstractmethod
import numpy as np
import trimesh
from shapely.geometry import MultiPolygon
from altprint.height_method import HeightMethod

class SlicedPlanes:
    """Represents the section planes obtained from the slicing of an object"""

    _height = float
    _planes_dict = dict[_height, MultiPolygon]
    _coord = tuple[float, float, float]
    _bounds_coords = tuple[_coord, _coord]

    def __init__(self, planes: _planes_dict, bounds : _bounds_coords):

        self.planes = planes
        self.bounds = bounds

    def get_heights(self):
        return list(self.planes.keys())


class Slicer(ABC):
    """Slicer base object"""

    @abstractmethod
    def load_model(self, model_file: str):
        pass

    @abstractmethod
    def translate_model(self, translation):
        pass

    @abstractmethod
    def slice_model(self) -> SlicedPlanes:
        pass

class STLSlicer(Slicer):
    """Slice .stl cad files"""

    def __init__(self, height_method: HeightMethod):
        self.height_method = height_method

    def load_model(self, model_file: str):
        self.model = trimesh.load_mesh(model_file)

    def translate_model(self, translation):
        self.model.apply_translation(translation)

    def slice_model(self, heights = None) -> SlicedPlanes:
        if not heights:
            heights = self.height_method.get_heights(self.model.bounds)
        sections = self.model.section_multiplane([0, 0, 0], [0, 0, 1], heights)
        planes = {}
        for i, section in enumerate(sections):
            if section:
                planes[heights[i]] = MultiPolygon(list(section.polygons_full))
            else:
                planes[heights[i]] = []

        return SlicedPlanes(planes, self.model.bounds)



from shapely.geometry import LineString
from altprint.printable.base import BasePrint
from altprint.flow import extrude, calculate
import numpy as np
import re

class GcodeExporter:

    def __init__(self, start_script = '', end_script = ''):
        self.gcode_content: list[str] = []
        self.head_x: float = 0.0
        self.head_y: float = 0.0
        self.min_jump: float = 1
        self.start_script_fname = start_script
        self.end_script_fname = end_script

    def segment(self, x, y, z, e, v) -> str:
        segment = []
        segment.append('; segment\n')
        segment.append('G92 E0.0000\n')
        segment.append('G1 F{0:.3f}\n'.format(v[0]))
        if z is not None:
            segment.append('G1 Z{0:.3f}\n'.format(z))
        segment.append('G1 X{0:.3f} Y{1:.3f}\n'.format(x[0], y[0]))

        actual_speed = v[0]
        for i in range(len(x)-1):
            if actual_speed != v[i+1]:
                segment.append('G1 X{0:.3f} Y{1:.3f} E{2:.4f} F{1:.3f} \n'.format(x[i+1], y[i+1], e[i+1], v[i+1]))
                actual_speed = v[i+1]
            else:
                segment.append('G1 X{0:.3f} Y{1:.3f} E{2:.4f} \n'.format(x[i+1], y[i+1], e[i+1]))
        segment.append('G92 E0.0000\n')
        segment = "".join(segment)
        return segment

    def jump(self, x, y, v=12000) -> str:
        jump = []
        jump.append('; jumping\n')
        jump.append('G92 E3.0000\n')
        jump.append('G1 E0 F2400\n')
        jump.append('G1 X{0:.3f} Y{1:.3f} F{2:.3f}\n'.format(x, y, v))
        jump.append('G1 E3 F2400\n')
        jump.append('G92 E0.0000\n')
        jump = "".join(jump)
        return jump


    def read_script(self, fname):
        script = ""
        with open(fname, 'r') as f:
            script = f.readlines()
            script = ''.join(script)
        return script

    def make_gcode(self, printable: BasePrint):

        self.gcode_content = []
        start_script = self.read_script(self.start_script_fname)
        end_script = self.read_script(self.end_script_fname)
        self.gcode_content.append(start_script)

        for z, layer in printable.layers.items():
            for raster in layer.perimeter:
                x, y = raster.path.xy
                x, y = np.array(x), np.array(y)
                if LineString([(self.head_x, self.head_y), (x[0], y[0])]).length > self.min_jump:
                    self.gcode_content.append(self.jump(x[0], y[0]))
                self.head_x, self.head_y = x[-1], y[-1]
                self.gcode_content.append(self.segment(x, y, z, raster.extrusion, raster.speed))

            for raster in layer.infill:
                x, y = raster.path.xy
                x, y = np.array(x), np.array(y)
                if LineString([(self.head_x, self.head_y), (x[0], y[0])]).length > self.min_jump:
                    self.gcode_content.append(self.jump(x[0], y[0]))
                self.head_x, self.head_y = x[-1], y[-1]
                self.gcode_content.append(self.segment(x, y, z, raster.extrusion, raster.speed))

        self.gcode_content.append(end_script)

    def make_layer_gcode(self, layer):
        layer_gcode = []
        for raster in layer.perimeter:
            x, y = raster.path.xy
            x, y = np.array(x), np.array(y)
            if LineString([(self.head_x, self.head_y), (x[0], y[0])]).length > self.min_jump:
                layer_gcode.append(self.jump(x[0], y[0]))
            self.head_x, self.head_y = x[-1], y[-1]
            layer_gcode.append(self.segment(x, y, None, raster.extrusion, raster.speed))

        for raster in layer.infill:
            x, y = raster.path.xy
            x, y = np.array(x), np.array(y)
            if LineString([(self.head_x, self.head_y), (x[0], y[0])]).length > self.min_jump:
                layer_gcode.append(self.jump(x[0], y[0]))
            self.head_x, self.head_y = x[-1], y[-1]
            layer_gcode.append(self.segment(x, y, None, raster.extrusion, raster.speed))

        return layer_gcode

    def export_gcode(self, filename):
        with open(filename, 'w') as f:
            for gcode_block in self.gcode_content:
                f.write(gcode_block)
